import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

import '../../osam_flutter.dart';

PageRoute _pageRoute({WidgetBuilder widgetBuilder, RouteSettings settings, bool isMaterial = false}) => isMaterial
    ? MaterialPageRoute(builder: widgetBuilder, settings: settings)
    : CupertinoPageRoute(builder: widgetBuilder, settings: settings);

class OsamNavigator<P extends NavigationPresenter> extends StatefulWidget {
  final Map<String, dynamic> routes;
  final bool isMaterial;
  final bool log;

  const OsamNavigator({Key key, this.routes, this.isMaterial = false, this.log = false}) : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState<P>();
}

class _OsamNavigatorState<P extends NavigationPresenter> extends State<OsamNavigator> with AfterLayoutMixin {
  final List<String> _routes = [];
  var lastKnownStackSize = 0;

  Map<String, dynamic> get routes => widget.routes;

  @override
  Widget build(BuildContext context) => routes[context.presenter<P>().routesStream.value.first](context);

  void _silentPop(NavigationPresenter presenter) {
    if (_routes.length > 1) {
      lastKnownStackSize = lastKnownStackSize - 1;
      if (_routes.length > lastKnownStackSize) {
        _routes.removeLast();
        presenter.pop();
      }
    }
  }

  bool isWidgetBuilder(String route) => routes[route] is WidgetBuilder;

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context);
    final presenter = context.presenter<P>();
    final initialRoutes = presenter.routesStream.value;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        final target = routes[route];
        if (!isWidgetBuilder(route)) {
          (target() as Future).then((_) {
            _silentPop(presenter);
          });
        } else {
          navigator.push(_pageRoute(widgetBuilder: target, isMaterial: widget.isMaterial)
            ..popped.then((_) => _silentPop(presenter)));
        }
      });
    }
    lastKnownStackSize = _routes.length;
    presenter.routesSub.onData((newRoutes) {
      if (widget.log) {
        print('Last routes: $_routes');
        print('New routes: $newRoutes');
      }
      final routesDifferentNumber = newRoutes.length - _routes.length;
      if (routesDifferentNumber == 0 && newRoutes.last != _routes.last) {
        final target = routes[newRoutes.last];
        if (!isWidgetBuilder(newRoutes.last)) {
          (target() as Future).then((_) {
            _silentPop(presenter);
          });
        } else {
          navigator.pushReplacement(_pageRoute(widgetBuilder: target, isMaterial: widget.isMaterial)
            ..popped.then((_) => _silentPop(presenter)));
        }
        lastKnownStackSize = lastKnownStackSize + 1;
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = newRoutes.sublist(_routes.length, newRoutes.length);
        pushRoutes.forEach((route) {
          if (!isWidgetBuilder(route)) {
            (routes[route]() as Future).then((_) {
              _silentPop(presenter);
            });
          } else {
            navigator.push(_pageRoute(widgetBuilder: routes[route], isMaterial: widget.isMaterial)
              ..popped.then((_) {
                _silentPop(presenter);
              }));
          }
        });
        lastKnownStackSize = newRoutes.length;
      } else if (routesDifferentNumber < 0) {
        if (_routes.first != newRoutes.first) {
          lastKnownStackSize = newRoutes.length;
          // unsupported for dialogs and others
          if (!isWidgetBuilder(newRoutes.first)) throw Error();
          navigator.pushAndRemoveUntil(
              _pageRoute(widgetBuilder: routes[newRoutes.first], isMaterial: widget.isMaterial), (route) => false);
        } else {
          _routes.sublist(newRoutes.length, _routes.length).forEach((route) {
            navigator.pop();
          });
        }
      }
      _routes.clear();
      _routes.addAll(newRoutes);
    });
  }
}

class OsamTabNavigator<P extends TabNavigationPresenter> extends StatelessWidget {
  final Map<String, PresenterProvider<NavigationPresenter>> pages;
  final Map<String, Widget> tabs;
  final BoxDecoration tabDecoration;
  final Map<String, GlobalKey<NavigatorState>> navigatorKeys;

  const OsamTabNavigator({Key key, this.pages, this.tabDecoration, this.tabs, this.navigatorKeys}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final presenter = context.presenter<P>();
    return OsamBuilder<String>(
        stream: presenter.tabStream,
        builder: (ctx, value) {
          final currentPresenter = pages[value].presenter;
          return Stack(
            children: <Widget>[
              IndexedStack(
                index: pages.keys.toList().indexOf(value),
                children: pages.keys.map((tab) => pages[tab]).toList(),
              ),
              Positioned(
                bottom: 0,
                child: SafeArea(
                  bottom: true,
                  child: Container(
                    decoration: tabDecoration,
                    padding: const EdgeInsets.symmetric(vertical: 5),
                    constraints: BoxConstraints(
                      maxWidth: MediaQuery.of(context).size.width,
                      minHeight: 48,
                    ),
                    child: Row(
                      children: pages.keys
                          .map((tab) => Expanded(
                                child: GestureDetector(
                                  child: Container(
                                    child: tabs[tab],
                                    alignment: Alignment.center,
                                    color: Colors.transparent,
                                  ),
                                  onTap: () {
                                    if (value != tab) {
                                      presenter.changeTab(tab);
                                    } else {
                                      final routes = currentPresenter.routesStream.value;
                                      if (routes.length > 1) {
                                        currentPresenter.pushAndRemoveUntil(routes.first);
                                      }
                                    }
                                  },
                                ),
                              ))
                          .toList(),
                    ),
                  ),
                ),
              )
            ],
          );
        });
  }
}
